/*
 * XPSDriver.cpp
 * Asyn driver that controls the analog and digital I/O on a Newport XPS 
 *
 * Author: Mark Rivers
 *
 * Created January 25, 2015
 */

#include <string.h>
#include <stdlib.h>

#include <epicsThread.h>
#include <epicsString.h>
#include <epicsExit.h>
#include <asynOctetSyncIO.h>
#include <iocsh.h>

#include <epicsExport.h>

#include "XPSDriver.h"

#define XPS_TIMEOUT 1.0
#define MAX_ANALOG_INPUTS   4
#define MAX_ANALOG_OUTPUTS  4
#define MAX_ADDRESS MAX_ANALOG_INPUTS

static const char *analogInputNames[MAX_ANALOG_INPUTS] = {
    "GPIO2.ADC1", /* Analog Input # 1 of the I/O board connector # 2 */
    "GPIO2.ADC2", /* Analog Input # 2 of the I/O board connector # 2 */
    "GPIO2.ADC3", /* Analog Input # 3 of the I/O board connector # 2 */
    "GPIO2.ADC4", /* Analog Input # 4 of the I/O board connector # 2 */
};
static const char *analogOutputNames[MAX_ANALOG_OUTPUTS] = {
    "GPIO2.DAC1", /* Analog Output # 1 of the I/O board connector # 2 */
    "GPIO2.DAC2", /* Analog Output # 2 of the I/O board connector # 2 */
    "GPIO2.DAC3", /* Analog Output # 3 of the I/O board connector # 2 */
    "GPIO2.DAC4", /* Analog Output # 4 of the I/O board connector # 2 */
};

static const char *digitalInputName  = "GPIO3.DI";
static const char *digitalOutputName = "GPIO3.DO";

static const char *driverName="XPSDriver";
static void pollerThread(void *drvPvt);
static void exitHandler(void *drvPvt);

/** Constructor for the XPSDriver class.
  * Calls constructor for the asynPortDriver base class.
  * \param[in] portName The name of the asyn port driver to be created.
  * \param[in] XPSName The asyn port name of the drvAsynIPPort connect to the XPS
  */
XPSDriver::XPSDriver(const char *portName, const char *XPSPortName, double pollDelay) 
   : asynPortDriver(portName, 
                    MAX_ADDRESS,   /* maxAddr */ 
                    NUM_XPS_PARAMS,
                    asynFloat64Mask | asynUInt32DigitalMask | asynOctetMask | asynDrvUserMask, /* Interface mask */
                    asynFloat64Mask | asynUInt32DigitalMask | asynOctetMask,                   /* Interrupt mask */
                    ASYN_CANBLOCK | ASYN_MULTIDEVICE, /* asynFlags.  This driver blocks it is multi-device */
                    1, /* Autoconnect */
                    0, /* Default priority */
                    0) /* Default stack size*/    
{
    asynStatus status;
    const char *functionName = "XPSDriver";
    
    // Create the parameters for this driver
    createParam(P_AnalogInputString,        asynParamFloat64,       &P_AnalogInput);
    createParam(P_AnalogOutputString,       asynParamFloat64,       &P_AnalogOutput);
    createParam(P_DigitalInputString,       asynParamUInt32Digital, &P_DigitalInput);
    createParam(P_DigitalOutputString,      asynParamUInt32Digital, &P_DigitalOutput);
    createParam(P_FirmwareVersionString,    asynParamOctet,         &P_FirmwareVersion);
    
    XPSPortName_ = epicsStrDup(XPSPortName);
    pollDelay_ = pollDelay;
    exiting_ = false;

    // Connect to the server
    status = pasynOctetSyncIO->connect(XPSPortName, 0, &pasynUserXPS_, NULL);
    if (status) {
        printf("%s:%s: error calling pasynOctetSyncIO->connect, status=%d, error=%s\n", 
               driverName, functionName, status, pasynUserXPS_->errorMessage);
        return;
    }
        
    // Read the firmware version
    sprintf(outString_, "FirmwareVersionGet(char *)");
    status = writeReadXPS();
    if (status) return;
    setStringParam(P_FirmwareVersion, inString_);
    
    // Create the thread that polls the XPS
    status = (asynStatus)(epicsThreadCreate("XPSDriverTask",
                          epicsThreadPriorityMedium,
                          epicsThreadGetStackSize(epicsThreadStackMedium),
                          (EPICSTHREADFUNC)::pollerThread,
                          this) == NULL);
    if (status) {
        printf("%s:%s: epicsThreadCreate failure, status=%d\n", driverName, functionName, status);
        return;
    }

    // Establish an exit handler
    epicsAtExit(::exitHandler, this);
    
}

/** Called when asyn clients call pasynUInt32Digital->write().
  * \param[in] pasynUser pasynUser structure that encodes the reason and address.
    * \param[in] mask Mask value for write operation
  * \param[in] value Value to write. */
asynStatus XPSDriver::writeUInt32Digital(asynUser *pasynUser, epicsUInt32 value, epicsUInt32 mask)
{
    int function = pasynUser->reason;
    int status = asynSuccess;
    int channel;
    const char *paramName;
    const char* functionName = "writeUInt32Digital";

    getAddress(pasynUser, &channel);
    
    /* Set the parameter in the parameter library. */
    status |= setUIntDigitalParam(channel, function, value, mask);
    
    /* Fetch the parameter string name for possible use in debugging */
    getParamName(function, &paramName);

    if (function == P_DigitalOutput) {
        sprintf(outString_, "GPIODigitalSet(%s,%hu,%hu)", digitalOutputName, mask, value);
        status |= writeReadXPS();
    } 
    else {
        /* All other parameters just get set in parameter list, no need to
         * act on them here */
    }
    
    /* Do callbacks so higher layers see any changes */
    status |= (asynStatus) callParamCallbacks();
    
    if (status) 
        epicsSnprintf(pasynUser->errorMessage, pasynUser->errorMessageSize, 
            "%s:%s: status=%d, function=%d, name=%s, value=%d", 
            driverName, functionName, status, function, paramName, value);
    else        
        asynPrint(pasynUser, ASYN_TRACEIO_DRIVER, 
            "%s:%s: function=%d, name=%s, value=%d\n", 
            driverName, functionName, function, paramName, value);
    return (asynStatus)status;
}

/** Called when asyn clients call pasynFloat64->write().
  * This function sends a signal to the simTask thread if the value of P_UpdateTime has changed.
  * For all  parameters it  sets the value in the parameter library and calls any registered callbacks.
  * \param[in] pasynUser pasynUser structure that encodes the reason and address.
  * \param[in] value Value to write. */
asynStatus XPSDriver::writeFloat64(asynUser *pasynUser, epicsFloat64 value)
{
    int function = pasynUser->reason;
    int status = asynSuccess;
    int channel;
    const char *paramName;
    const char* functionName = "writeFloat64";

    getAddress(pasynUser, &channel);
    
    /* Set the parameter in the parameter library. */
    status |= setDoubleParam(channel, function, value);

    /* Fetch the parameter string name for possible use in debugging */
    getParamName(function, &paramName);

    if (function == P_AnalogOutput) {
        sprintf(outString_, "GPIOAnalogSet(%s,%f)", analogOutputNames[channel], value);
        status |= writeReadXPS();
    } else {
        /* All other parameters just get set in parameter list, no need to
         * act on them here */
    }
    
    /* Do callbacks so higher layers see any changes */
    status |= (asynStatus) callParamCallbacks();
    
    if (status) 
        epicsSnprintf(pasynUser->errorMessage, pasynUser->errorMessageSize, 
            "%s:%s: status=%d, function=%d, name=%s, value=%f", 
            driverName, functionName, status, function, paramName, value);
    else        
        asynPrint(pasynUser, ASYN_TRACEIO_DRIVER, 
            "%s:%s: function=%d, name=%s, value=%f\n", 
            driverName, functionName, function, paramName, value);
    return (asynStatus)status;
}

/** Writes a string to the XPS and reads the response. */
asynStatus XPSDriver::writeReadXPS()
{
    size_t nread;
    size_t nwrite;
    asynStatus status;
    bool replyOK;
    char *eom;
    int eomReason;
    int len;
    char temp[MAX_MESSAGE_SIZE];
    const char *functionName="writeReadXPS";
  
    status = pasynOctetSyncIO->writeRead(pasynUserXPS_, outString_, strlen(outString_), 
                                         temp, sizeof(temp), XPS_TIMEOUT, &nwrite, &nread, &eomReason);
    replyOK = strncmp(temp, "0,", 2) == 0;
    eom = strstr(temp, ",EndOfAPI");
    if (status || !replyOK || eom==0) {
        asynPrint(pasynUserSelf, ASYN_TRACE_ERROR, 
            "%s::%s error in communication with XPS, status=%d, output=%s, input=%s, nwrite=%d, nread=%d, eomReason=%d\n",
            driverName, functionName, status, outString_, temp, (int)nwrite, (int)nread, eomReason);
        return asynError;
    }
    len = eom - temp - 2;
    strncpy(inString_, &temp[2], len);
    inString_[len] = 0;
    asynPrint(pasynUserSelf, ASYN_TRACEIO_DRIVER, 
            "%s::%s successful communication with XPS, output=%s, raw input=%s, trimmed input=%s, nwrite=%d, nread=%d, len=%d, eomReason=%d\n",
            driverName, functionName, outString_, temp, inString_, (int)nwrite, (int)nread, len, eomReason);
  return status;
}


static void pollerThread(void *drvPvt)
{
    XPSDriver *pPvt = (XPSDriver *)drvPvt;
    
    pPvt->pollerThread();
}

/** Read thread to read the data from the electrometer when it is in continuous acquire mode.
  * Reads the data and does callbacks.
  */
void XPSDriver::pollerThread(void)
{
    int status;
    int i;
    double aiValue;
    epicsUInt32 uiValue;
    //static const char *functionName = "readThread";

    /* Loop forever */
    lock();
    while (1) {
        // Read each of the analog inputs
        for (i=0; i<MAX_ANALOG_INPUTS; i++) {
            sprintf(outString_, "GPIOAnalogGet(%s,double *)", analogInputNames[i]);
            status = writeReadXPS();
            if (status == asynSuccess) {
                aiValue = atof(inString_);
                setDoubleParam(i, P_AnalogInput, aiValue);
                callParamCallbacks(i);
            }
        }
        // Read each of the analog outputs
        for (i=0; i<MAX_ANALOG_INPUTS; i++) {
            sprintf(outString_, "GPIOAnalogGet(%s,double *)", analogOutputNames[i]);
            status = writeReadXPS();
            if (status == asynSuccess) {
                aiValue = atof(inString_);
                setDoubleParam(i, P_AnalogOutput, aiValue);
                callParamCallbacks(i);
            }
        }
        // Read the digital input register
        sprintf(outString_, "GPIODigitalGet(%s,unsigned short *)", digitalInputName);
        status = writeReadXPS();
        if (status == asynSuccess) {
            uiValue = atoi(inString_);
            setUIntDigitalParam(P_DigitalInput, uiValue, 0xffffffff);
            callParamCallbacks();
        }
        // Read the digital output register
        sprintf(outString_, "GPIODigitalGet(%s,unsigned short *)", digitalOutputName);
        status = writeReadXPS();
        if (status == asynSuccess) {
            uiValue = atoi(inString_);
            setUIntDigitalParam(P_DigitalOutput, uiValue, 0xffffffff);
            callParamCallbacks();
        }
        unlock();
        epicsThreadSleep(pollDelay_);
        lock();
        if (exiting_) break;
    }
}

/** Report  parameters 
  * \param[in] fp The file pointer to write to
  * \param[in] details The level of detail requested
  */
void XPSDriver::report(FILE *fp, int details)
{
    char firmwareVersion[MAX_MESSAGE_SIZE];
        
    getStringParam(P_FirmwareVersion, sizeof(firmwareVersion), firmwareVersion);
    fprintf(fp, "%s: port=%s, XPS port=%s, firmware version=%s, pollDelay=%f\n",
            driverName, portName, XPSPortName_, firmwareVersion, pollDelay_);
    asynPortDriver::report(fp, details);
}

static void exitHandler(void *drvPvt)
{
    XPSDriver *pPvt = (XPSDriver *)drvPvt;
    
    pPvt->exitHandler();
}

void XPSDriver::exitHandler()
{
    exiting_ = true;
}


/* Configuration routine.  Called directly, or from the iocsh function below */

extern "C" {

/** EPICS iocsh callable function to call constructor for the XPSDriver class.
  * \param[in] portName The name of the asyn port driver to be created.
  * \param[in] XPSPortName The name of the asyn communication port to the server
  *            created with drvAsynIPPortConfigure */
int XPSDriverConfigure(const char *portName, const char *XPSPortName, double pollDelay)
{
    new XPSDriver(portName, XPSPortName, pollDelay);
    return asynSuccess;
}


/* EPICS iocsh shell commands */

static const iocshArg initArg0 = { "portName",iocshArgString};
static const iocshArg initArg1 = { "XPSPortName",iocshArgString};
static const iocshArg initArg2 = { "XPSPortName",iocshArgDouble};
static const iocshArg * const initArgs[] = {&initArg0,
                                            &initArg1,
                                            &initArg2};
static const iocshFuncDef initFuncDef = {"XPSDriverConfigure",3,initArgs};
static void initCallFunc(const iocshArgBuf *args)
{
    XPSDriverConfigure(args[0].sval, args[1].sval, args[2].dval);
}

void XPSDriverRegister(void)
{
    iocshRegister(&initFuncDef,initCallFunc);
}

epicsExportRegistrar(XPSDriverRegister);

}  /* extern "C" */

